import java.util.*;

public class Espressione {
	private Integer number;

	private String op;

	private Espressione e1, e2;

	public Espressione(Integer number) {
		this.number = number;
	}

	public Espressione(String op, Espressione e1, Espressione e2) {
		this.op = op;
		this.e1 = e1;
		this.e2 = e2;
	}

	public Integer valuta() {
		if ("+".equals(op)) return e1.valuta() + e2.valuta();
		if ("*".equals(op)) return e1.valuta() * e2.valuta();
		if ("-".equals(op)) return e1.valuta() - e2.valuta();
		return number;
	}

	public String toString() {
		if (op == null) return number.toString();
		return "(" + e1 + op + e2 + ")";
	}

	public static Espressione parseDiretta(Iterator<Object> it) {
		Object token = it.next();
		if (token instanceof Integer) return new Espressione((Integer) token);
		return new Espressione((String) token, parseDiretta(it), parseDiretta(it));
	}

	public static Espressione parseInversa(Iterator<Object> it) {
		Stack<Espressione> stack = new Stack<Espressione>();
		while (it.hasNext()) {
			Object token = it.next();
			if (token instanceof Integer) {
				stack.push(new Espressione((Integer) token));
			} else {
				Espressione e2 = stack.pop();
				Espressione e1 = stack.pop();
				stack.push(new Espressione((String) token, e1, e2));
			}
		}
		return stack.pop();
	}

	public static Espressione parseInfissa(Iterator<Object> it) {
		Object token = it.next();
		if (token instanceof Integer) return new Espressione((Integer) token);
		Espressione e = parseInfissa(it);
		e = new Espressione((String) it.next(), e, parseInfissa(it));
		it.next();
		return e;
	}

	static class Scan implements Iterator<Object> {
		private String s;

		private int i = 0;

		public Scan(String s) {
			this.s = s;
		}

		public boolean hasNext() {
			return i < s.length();
		}

		public Object next() {
			String c = "" + s.charAt(i++);
			try {
				return Integer.parseInt(c);
			}
			catch (NumberFormatException e) {
				return c;
			}
		}

		public void remove() {
			throw new UnsupportedOperationException();
		}

	}

	public static void main(String[] args) {
		Iterator<Object> it1 = new Espressione.Scan("*+89-53");
		Espressione e1 = Espressione.parseDiretta(it1);
		System.out.println(e1 + "=" + e1.valuta());

		Iterator<Object> it2 = new Espressione.Scan("((8+9)*(5-3))");
		Espressione e2 = Espressione.parseInfissa(it2);
		System.out.println(e2 + "=" + e2.valuta());

		Iterator<Object> it3 = new Espressione.Scan("98+53-*");
		Espressione e3 = Espressione.parseInversa(it3);
		System.out.println(e3 + "=" + e3.valuta());
	}

}
